from django.conf import settings
from django.db import models
from django.utils.translation import gettext_lazy as _
from django.utils import timezone
from django.contrib.contenttypes.models import ContentType
from django.contrib.postgres.fields import DateTimeRangeField
from oscar.models.fields import NullCharField
from . import settings as app_settings
from .utils import GeneratorRepository
import uuid
import os.path
import io


def get_report_upload_path(instance, filename):
    # Upload files to {MEDIA_ROOT}/{OSCAR_REPORTS_UPLOAD_PREFIX}/{YYYY}/{MM}/{DD}/{uuid}.{ext}
    extension = os.path.splitext(filename)[1].replace('.', '')
    return '{prefix}/{date}/{uuid}.{ext}'.format(
        prefix=app_settings.OSCAR_REPORTS_UPLOAD_PREFIX,
        date=instance.created_on.strftime('%Y/%m/%d'),
        uuid=instance.uuid,
        ext=extension)


class Report(models.Model):
    STATUS_CREATED = 'created'
    STATUS_QUEUED = 'queued'
    STATUS_IN_PROGRESS = 'in-progress'
    STATUS_COMPLETED = 'completed'
    STATUS_NAMES = {
        STATUS_CREATED: _("Created"),
        STATUS_QUEUED: _("Queued"),
        STATUS_IN_PROGRESS: _("In-Progress"),
        STATUS_COMPLETED: _("Completed"),
    }

    # Unique ID used in URLs and filenames
    uuid = models.UUIDField(_("UUID"), default=uuid.uuid4, editable=False, unique=True)

    # Store a link to the content-type (model) which was used to generate this report.
    # Only user's with the `can_view` permission on this content-type will be able to
    # see this report.
    content_type = models.ForeignKey(ContentType,
        verbose_name=_("Content Type"),
        on_delete=models.CASCADE,
        related_name='+',
        null=True,
        blank=True)

    # Report Metadata
    owner = models.ForeignKey(settings.AUTH_USER_MODEL,
        verbose_name=_("Owner"),
        null=True,
        on_delete=models.SET_NULL,
        related_name='+')
    type_code = models.CharField(_("Type Code"), max_length=50)
    description = models.TextField(_("Description"))
    date_range = DateTimeRangeField(_("Date Range"), null=True, blank=True)

    # Celery Task ID
    task_id = models.UUIDField(_("Background Task ID"), editable=False, unique=True, null=True, blank=True)

    # Status Timestamps
    created_on = models.DateTimeField(_("Created On"), auto_now_add=True)
    queued_on = models.DateTimeField(_("Queued On"), null=True, blank=True)
    started_on = models.DateTimeField(_("Started On"), null=True, blank=True)
    completed_on = models.DateTimeField(_("Completed On"), null=True, blank=True)

    # Report File Output
    mime_type = NullCharField(_("MIME Type"), max_length=20)
    report_file = models.FileField(_("Report File"),
        upload_to=get_report_upload_path,
        null=True,
        blank=True)


    def __str__(self):
        return str(self.uuid)


    @property
    def status(self):
        if self.completed_on:
            return self.STATUS_COMPLETED
        if self.started_on:
            return self.STATUS_IN_PROGRESS
        if self.queued_on:
            return self.STATUS_QUEUED
        return self.STATUS_CREATED


    @property
    def status_name(self):
        return self.STATUS_NAMES.get(self.status, "")


    @property
    def is_complete(self):
        return self.status == self.STATUS_COMPLETED


    @property
    def generator_class(self):
        generator = GeneratorRepository().get_generator(self.type_code)
        if not generator:
            raise ValueError("Can not queue report: no generator class was found.")
        return generator


    def queue(self, delay=10, report_format='CSV'):
        # Reset metadata
        self.queued_on = None
        self.started_on = None
        self.completed_on = None
        self.task_id = None
        self.save(update_fields=[
            'queued_on',
            'started_on',
            'completed_on',
            'task_id',
        ])
        # Queue the task using celery
        generator = self.get_generator(report_format)
        from . import tasks
        task = tasks.generate_report.apply_async(args=[self.uuid, report_format], countdown=delay)
        # Update the task metadata
        self.description = generator.report_description()
        self.queued_on = timezone.now()
        self.task_id = task.id
        self.save(update_fields=[
            'description',
            'queued_on',
            'task_id',
        ])
        return task
    queue.alters_data = True


    def generate(self, report_format='CSV'):
        # Record start time
        self.started_on = timezone.now()
        self.save(update_fields=['started_on'])
        # Generate report content
        generator = self.get_generator(report_format)
        report = generator.generate()
        # Save generated content
        filename = self.get_filename(report_format)
        content = io.BytesIO(report.content)
        self.mime_type = report['content-type']
        self.report_file.save(filename, content)
        self.completed_on = timezone.now()
        self.save(update_fields=[
            'mime_type',
            'report_file',
            'completed_on',
        ])
    generate.alters_data = True


    def get_generator(self, report_format):
        kwargs = {
            'start_date': self.date_range.lower if self.date_range else None,
            'end_date': self.date_range.upper if self.date_range else None,
            'formatter': report_format,
        }
        return self.generator_class(**kwargs)


    def get_filename(self, report_format):
        return '{}.{}'.format(self.uuid, report_format.lower())



from django.urls import reverse
from oscar.test.testcases import WebTestCase


class ReportsDashboardTests(WebTestCase):
    is_staff = True


    def test_dashboard_is_accessible_to_staff(self):
        url = reverse('dashboard:reports-index')
        response = self.get(url)
        self.assertIsOk(response)


    def test_conditional_offers_no_date_range(self):
        url = reverse('dashboard:reports-index')
        response = self.get(url)

        response.form['report_type'] = 'conditional-offers'
        response.form.submit()
        self.assertIsOk(response)


    def test_conditional_offers_with_date_range(self):
        url = reverse('dashboard:reports-index')
        response = self.get(url)

        response.form['report_type'] = 'conditional-offers'
        response.form['date_from'] = '2017-01-01'
        response.form['date_to'] = '2017-12-31'
        response.form.submit()
        self.assertIsOk(response)


    def test_conditional_offers_with_invalid_date_range(self):
        url = reverse('dashboard:reports-index')
        response = self.get(url)

        response.form['report_type'] = 'conditional-offers'
        response.form['date_from'] = '2017-12-31'
        response.form['date_to'] = '2017-01-01'
        response.form.submit()
        self.assertIsOk(response)



from django.contrib import admin
from . import models


@admin.register(models.Report)
class ReportAdmin(admin.ModelAdmin):
    search_fields = ['uuid']
    raw_id_fields = ['owner']
    list_display = [
        'uuid',
        'owner',
        'content_type',
        'type_code',
        'created_on',
        'queued_on',
        'started_on',
        'completed_on',
        'status_name',
    ]
    list_filter = [
        'type_code',
        'created_on',
        'queued_on',
        'started_on',
        'completed_on',
    ]
    readonly_fields = [
        'uuid',
        'task_id',
        'created_on',
    ]
    fields = [
        'uuid',
        'content_type',
        'type_code',
        'owner',
        'description',
        'date_range',
        'task_id',
        'created_on',
        'queued_on',
        'started_on',
        'completed_on',
        'mime_type',
        'report_file',
    ]



from datetime import timedelta
from django.test import TestCase
from django.contrib.auth.models import User
from django.utils import timezone
from freezegun import freeze_time
from psycopg2.extras import DateTimeTZRange
from unittest import mock
from .. import models


@freeze_time('2019-10-03T12:00:00-04:00')
class ReportTest(TestCase):

    def setUp(self):
        self.staff_user = User.objects.create_user(
            username='root',
            is_staff=True)

        self.report = models.Report()
        self.report.uuid = 'd3c74a8b-e7ae-4482-bd9c-bee69fde5c5c'
        self.report.content_type = None
        self.report.owner = self.staff_user
        self.report.type_code = 'order_report'
        self.report.date_range = DateTimeTZRange(
            lower=(timezone.now() - timedelta(days=1)),
            upper=(timezone.now()))
        self.report.save()


    def test_str(self):
        self.assertEqual(str(self.report), 'd3c74a8b-e7ae-4482-bd9c-bee69fde5c5c')


    def test_status(self):
        self.report.queued_on = None
        self.report.started_on = None
        self.report.completed_on = None
        self.report.save()
        self.assertEqual(self.report.status, 'created')
        self.assertEqual(self.report.status_name, 'Created')
        self.assertFalse(self.report.is_complete)

        self.report.queued_on = timezone.now()
        self.report.save()
        self.assertEqual(self.report.status, 'queued')
        self.assertEqual(self.report.status_name, 'Queued')
        self.assertFalse(self.report.is_complete)

        self.report.started_on = timezone.now()
        self.report.save()
        self.assertEqual(self.report.status, 'in-progress')
        self.assertEqual(self.report.status_name, 'In-Progress')
        self.assertFalse(self.report.is_complete)

        self.report.completed_on = timezone.now()
        self.report.save()
        self.assertEqual(self.report.status, 'completed')
        self.assertEqual(self.report.status_name, 'Completed')
        self.assertTrue(self.report.is_complete)


    def test_generator_class(self):
        self.assertEqual(self.report.generator_class.description, 'Orders placed')


    def test_generator_class_missing(self):
        self.report.type_code = 'foo_bar_baz'
        with self.assertRaises(ValueError):
            print(self.report.generator_class)


    @mock.patch('oscarreports.tasks.generate_report')
    def test_queue(self, mock_generate_report):
        task = mock.MagicMock()
        task.id = 'f3a0b0a0-148a-4ba1-9ed4-dd5543e77d73'
        mock_generate_report.apply_async.return_value = task

        self.assertEqual(mock_generate_report.apply_async.call_count, 0)

        self.assertEqual(self.report.description, '')
        self.assertIsNone(self.report.queued_on)
        self.assertIsNone(self.report.started_on)
        self.assertIsNone(self.report.completed_on)
        self.assertIsNone(self.report.task_id)

        self.report.queue()

        self.assertEqual(mock_generate_report.apply_async.call_count, 1)
        mock_generate_report.apply_async.assert_called_once_with(
            args=['d3c74a8b-e7ae-4482-bd9c-bee69fde5c5c', 'CSV'],
            countdown=10)

        self.assertTrue(self.report.description.startswith('Orders placed between'))
        self.assertIsNotNone(self.report.queued_on)
        self.assertIsNone(self.report.started_on)
        self.assertIsNone(self.report.completed_on)
        self.assertEqual(self.report.task_id, 'f3a0b0a0-148a-4ba1-9ed4-dd5543e77d73')


    def test_generate(self):
        self.assertIsNone(self.report.started_on)
        self.assertIsNone(self.report.mime_type)
        self.assertIsNone(self.report.report_file.name)
        self.assertIsNone(self.report.completed_on)

        self.report.generate()

        self.assertIsNotNone(self.report.started_on)
        self.assertEqual(self.report.mime_type, 'text/csv')
        self.assertTrue(self.report.report_file.name.startswith('oscar-reports/2019/10/03/d3c74a8b-e7ae-4482-bd9c-bee69fde5c5c'))
        self.assertTrue(self.report.report_file.name.endswith('.csv'))
        self.assertIsNotNone(self.report.completed_on)


    def test_queue_integration(self):
        self.assertEqual(self.report.description, '')
        self.assertIsNone(self.report.queued_on)
        self.assertIsNone(self.report.started_on)
        self.assertIsNone(self.report.completed_on)
        self.assertIsNone(self.report.task_id)
        self.assertIsNone(self.report.mime_type)
        self.assertIsNone(self.report.report_file.name)

        task = self.report.queue(delay=0)
        self.assertIsNone(task.get())

        self.report.refresh_from_db()

        self.assertTrue(self.report.description.startswith('Orders placed between'))
        self.assertIsNotNone(self.report.queued_on)
        self.assertIsNotNone(self.report.started_on)
        self.assertIsNotNone(self.report.completed_on)
        self.assertEqual(str(self.report.task_id), task.id)
        self.assertEqual(self.report.mime_type, 'text/csv')
        self.assertTrue(self.report.report_file.name.startswith('oscar-reports/2019/10/03/d3c74a8b-e7ae-4482-bd9c-bee69fde5c5c'))
        self.assertTrue(self.report.report_file.name.endswith('.csv'))


    def test_delete(self):
        self.report.generate()
        self.report.report_file.delete = mock.MagicMock()

        self.assertEqual(self.report.report_file.delete.call_count, 0)

        self.report.delete()

        self.assertEqual(self.report.report_file.delete.call_count, 1)
        self.report.report_file.delete.assert_called_once_with(save=False)



from __future__ import absolute_import
from celery import shared_task



@shared_task(ignore_result=True)
def generate_report(report_uuid, report_format):
    from . import models
    report = models.Report.objects.get(uuid=report_uuid)
    report.generate(report_format)



# Generated by Django 2.2.6 on 2019-10-03 21:12

from django.conf import settings
import django.contrib.postgres.fields.ranges
from django.db import migrations, models
import django.db.models.deletion
import oscar.models.fields
import oscarreports.models
import uuid


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('contenttypes', '0002_remove_content_type_name'),
    ]

    operations = [
        migrations.CreateModel(
            name='Report',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('uuid', models.UUIDField(default=uuid.uuid4, editable=False, unique=True, verbose_name='UUID')),
                ('type_code', models.CharField(max_length=50, verbose_name='Type Code')),
                ('description', models.TextField(verbose_name='Description')),
                ('date_range', django.contrib.postgres.fields.ranges.DateTimeRangeField(blank=True, null=True, verbose_name='Date Range')),
                ('task_id', models.UUIDField(blank=True, editable=False, null=True, unique=True, verbose_name='Background Task ID')),
                ('created_on', models.DateTimeField(auto_now_add=True, verbose_name='Created On')),
                ('queued_on', models.DateTimeField(blank=True, null=True, verbose_name='Queued On')),
                ('started_on', models.DateTimeField(blank=True, null=True, verbose_name='Started On')),
                ('completed_on', models.DateTimeField(blank=True, null=True, verbose_name='Completed On')),
                ('mime_type', oscar.models.fields.NullCharField(max_length=20, verbose_name='MIME Type')),
                ('report_file', models.FileField(blank=True, null=True, upload_to=oscarreports.models.get_report_upload_path, verbose_name='Report File')),
                ('content_type', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='+', to='contenttypes.ContentType', verbose_name='Content Type')),
                ('owner', models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='+', to=settings.AUTH_USER_MODEL, verbose_name='Owner')),
            ],
        ),
    ]



#!/usr/bin/env python
from setuptools import setup, find_packages, Distribution
import codecs
import os.path

# Make sure versiontag exists before going any further
Distribution().fetch_build_eggs('versiontag>=1.2.0')

from versiontag import get_version, cache_git_tag  # NOQA


packages = find_packages('src')

install_requires = [
    'celery>=4.3',
    'django-oscar>=2.0.0',
]

extras_require = {
    'development': [
        'coverage>=4.4.2',
        'flake8>=3.5.0',
        'freezegun>=0.3.12',
        'psycopg2cffi>=2.7.7',
        'PyYAML>=3.12',
        'sorl-thumbnail>=11.04',
        'tox>=2.9.1',
        'versiontag>=1.2.0',

        # Needed for Oscar's web test framework
        'django-webtest==1.9.4',
        'WebTest>=2.0,<2.1',
    ],
}


def fpath(name):
    return os.path.join(os.path.dirname(__file__), name)


def read(fname):
    return codecs.open(fpath(fname), encoding='utf-8').read()


cache_git_tag()

setup(
    name='django-oscar-reports',
    description="An extension on-top of django-oscar that improves the Oscar Dashboard's report generation system",
    version=get_version(pypi=True),
    long_description=open('README.rst').read(),
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Web Environment',
        'Framework :: Django',
        'Framework :: Django :: 2.2',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: ISC License (ISCL)',
        'Operating System :: Unix',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],
    author='Craig Weber',
    author_email='crgwbr@gmail.com',
    url='https://gitlab.com/thelabnyc/django-oscar/django-oscar-reports',
    license='ISC',
    package_dir={'': 'src'},
    packages=packages,
    include_package_data=True,
    install_requires=install_requires,
    extras_require=extras_require,
)



from django.contrib import messages
from django.utils.translation import gettext_lazy as _
from django.shortcuts import redirect
from django.views.generic.edit import FormMixin
from django_tables2 import SingleTableView
from psycopg2.extras import DateTimeTZRange
from .models import Report
from .tables import ReportTable
from .forms import ReportForm


class IndexView(FormMixin, SingleTableView):
    template_name = 'oscar/dashboard/reports/index.html'
    table_pagination = True
    model = Report
    table_class = ReportTable
    context_table_name = 'reports'
    desc_template = ''
    description = ''

    form_class = ReportForm

    def dispatch(self, request, *args, **kwargs):
        form_class = self.get_form_class()
        self.form = self.get_form(form_class)
        return super().dispatch(request, *args, **kwargs)


    def get_queryset(self):
        queryset = self.model.objects.all().order_by('-created_on')
        return queryset


    def post(self, request, *args, **kwargs):
        if self.form.is_valid():
            # Create report
            report = Report()
            report.content_type = None
            report.owner = request.user
            report.type_code = self.form.cleaned_data['report_type']
            report.date_range = DateTimeTZRange(
                lower=self.form.cleaned_data['date_from'],
                upper=self.form.cleaned_data['date_to'])
            report.save()
            # Queue report
            report.queue()
            # Message user
            messages.info(request, _("Successfully queued report for generation"))
            return redirect('dashboard:reports-index')
        return self.get(request, *args, **kwargs)


    def get_table(self, **kwargs):
        table = super().get_table(**kwargs)
        table.caption = _("Reports")
        return table


    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context['form'] = self.form
        return context
